(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
  typeof define === 'function' && define.amd ? define(factory) :
  (global = global || self, global.AsyncComputed = factory());
}(this, function () { 'use strict';

  function isComputedLazy (item) {
    return item.hasOwnProperty('lazy') && item.lazy
  }

  function isLazyActive (vm, key) {
    return vm[lazyActivePrefix + key]
  }

  const lazyActivePrefix = 'async_computed$lazy_active$',
        lazyDataPrefix = 'async_computed$lazy_data$';

  function initLazy (data, key) {
    data[lazyActivePrefix + key] = false;
    data[lazyDataPrefix + key] = null;
  }

  function makeLazyComputed (key) {
    return {
      get () {
        this[lazyActivePrefix + key] = true;
        return this[lazyDataPrefix + key]
      },
      set (value) {
        this[lazyDataPrefix + key] = value;
      }
    }
  }

  function silentSetLazy (vm, key, value) {
    vm[lazyDataPrefix + key] = value;
  }
  function silentGetLazy (vm, key) {
    return vm[lazyDataPrefix + key]
  }

  const getGetterWatchedByArray = computedAsyncProperty =>
    function getter () {
      for (const key of computedAsyncProperty.watch) {
        // Check if nested key is watched.
        const splittedByDot = key.split('.');
        if (splittedByDot.length === 1) {
          // If not, just access it.
          // eslint-disable-next-line no-unused-expressions
          void this[key];
        } else {
          // Access the nested propety.
          try {
            let start = this;
            for (const part of splittedByDot) {
              start = start[part];
            }
          } catch (error) {
            console.error('AsyncComputed: bad path: ', key);
            throw error
          }
        }
      }
      return computedAsyncProperty.get.call(this)
    };

  const getGetterWatchedByFunction = computedAsyncProperty =>
    function getter () {
      computedAsyncProperty.watch.call(this);
      return computedAsyncProperty.get.call(this)
    };

  function getWatchedGetter (computedAsyncProperty) {
    if (typeof computedAsyncProperty.watch === 'function') {
      return getGetterWatchedByFunction(computedAsyncProperty)
    } else if (Array.isArray(computedAsyncProperty.watch)) {
      for (const key of computedAsyncProperty.watch) {
        if (typeof key !== 'string') {
          throw new Error('AsyncComputed: watch elemnts must be strings')
        }
      }
      return getGetterWatchedByArray(computedAsyncProperty)
    } else {
      throw Error('AsyncComputed: watch should be function or an array')
    }
  }

  const prefix = '_async_computed$';
  const DidNotUpdate = typeof Symbol === 'function' ? Symbol('did-not-update') : {};

  const AsyncComputed = {
    install (Vue, pluginOptions) {
      pluginOptions = pluginOptions || {};

      Vue.config
        .optionMergeStrategies
        .asyncComputed = Vue.config.optionMergeStrategies.computed;

      Vue.mixin({
        data () {
          return {
            _asyncComputed: {},
          }
        },
        beforeCreate () {
          const optionData = this.$options.data;
          const asyncComputed = this.$options.asyncComputed || {};

          if (!this.$options.computed) this.$options.computed = {};

          this.$options.computed.$asyncComputed = () => this.$data._asyncComputed;

          if (!Object.keys(asyncComputed).length) return

          for (const key in asyncComputed) {
            const getter = getterFn(key, this.$options.asyncComputed[key]);
            this.$options.computed[prefix + key] = getter;
          }

          this.$options.data = function vueAsyncComputedInjectedDataFn (vm) {
            const data = (
              (typeof optionData === 'function')
                ? optionData.call(this, vm)
                : optionData
            ) || {};
            for (const key in asyncComputed) {
              const item = this.$options.asyncComputed[key];
              if (isComputedLazy(item)) {
                initLazy(data, key);
                this.$options.computed[key] = makeLazyComputed(key);
              } else {
                data[key] = null;
              }
            }
            return data
          };
        },
        created () {
          for (const key in this.$options.asyncComputed || {}) {
            const item = this.$options.asyncComputed[key],
                  value = generateDefault.call(this, item, pluginOptions);
            if (isComputedLazy(item)) {
              silentSetLazy(this, key, value);
            } else {
              this[key] = value;
            }
          }

          for (const key in this.$options.asyncComputed || {}) {
            let promiseId = 0;
            const watcher = newPromise => {
              const thisPromise = ++promiseId;

              if (newPromise === DidNotUpdate) {
                return
              }

              if (!newPromise || !newPromise.then) {
                newPromise = Promise.resolve(newPromise);
              }
              setAsyncState(this, key, 'updating');

              newPromise.then(value => {
                if (thisPromise !== promiseId) return
                setAsyncState(this, key, 'success');
                this[key] = value;
              }).catch(err => {
                if (thisPromise !== promiseId) return

                setAsyncState(this, key, 'error');
                Vue.set(this.$data._asyncComputed[key], 'exception', err);
                if (pluginOptions.errorHandler === false) return

                const handler = (pluginOptions.errorHandler === undefined)
                  ? console.error.bind(console, 'Error evaluating async computed property:')
                  : pluginOptions.errorHandler;

                if (pluginOptions.useRawError) {
                  handler(err);
                } else {
                  handler(err.stack);
                }
              });
            };
            Vue.set(this.$data._asyncComputed, key, {
              exception: null,
              update: () => {
                watcher(getterOnly(this.$options.asyncComputed[key]).apply(this));
              }
            });
            setAsyncState(this, key, 'updating');
            this.$watch(prefix + key, watcher, { immediate: true });
          }
        }
      });
    }
  };

  function setAsyncState (vm, stateObject, state) {
    vm.$set(vm.$data._asyncComputed[stateObject], 'state', state);
    vm.$set(vm.$data._asyncComputed[stateObject], 'updating', state === 'updating');
    vm.$set(vm.$data._asyncComputed[stateObject], 'error', state === 'error');
    vm.$set(vm.$data._asyncComputed[stateObject], 'success', state === 'success');
  }

  function getterOnly (fn) {
    if (typeof fn === 'function') return fn

    return fn.get
  }

  function getterFn (key, fn) {
    if (typeof fn === 'function') return fn

    let getter = fn.get;

    if (fn.hasOwnProperty('watch')) {
      getter = getWatchedGetter(fn);
    }

    if (fn.hasOwnProperty('shouldUpdate')) {
      const previousGetter = getter;
      getter = function getter () {
        if (fn.shouldUpdate.call(this)) {
          return previousGetter.call(this)
        }
        return DidNotUpdate
      };
    }

    if (isComputedLazy(fn)) {
      const nonLazy = getter;
      getter = function lazyGetter () {
        if (isLazyActive(this, key)) {
          return nonLazy.call(this)
        } else {
          return silentGetLazy(this, key)
        }
      };
    }
    return getter
  }

  function generateDefault (fn, pluginOptions) {
    let defaultValue = null;

    if ('default' in fn) {
      defaultValue = fn.default;
    } else if ('default' in pluginOptions) {
      defaultValue = pluginOptions.default;
    }

    if (typeof defaultValue === 'function') {
      return defaultValue.call(this)
    } else {
      return defaultValue
    }
  }

  /* istanbul ignore if */
  if (typeof window !== 'undefined' && window.Vue) {
    // Auto install in dist mode
    window.Vue.use(AsyncComputed);
  }

  return AsyncComputed;

}));
